import { Collection, DropTarget, DropTargetDelegate, Key, Node } from "@react-types/shared";
import { InvalidationContext, Layout, LayoutInfo, Rect, Size } from "@react-stately/virtualizer";
import { GridNode } from "@react-types/grid";
import { TableCollection } from "@react-types/table";
export type ListLayoutOptions<T> = {
    /** The height of a row in px. */
    rowHeight?: number;
    estimatedRowHeight?: number;
    headingHeight?: number;
    estimatedHeadingHeight?: number;
    padding?: number;
    indentationForItem?: (collection: Collection<Node<T>>, key: Key) => number;
    loaderHeight?: number;
    placeholderHeight?: number;
    forceSectionHeaders?: boolean;
    enableEmptyState?: boolean;
};
export interface LayoutNode {
    node?: Node<unknown>;
    layoutInfo: LayoutInfo;
    header?: LayoutInfo;
    children?: LayoutNode[];
    validRect: Rect;
    index?: number;
}
export interface ListLayoutProps {
    isLoading?: boolean;
}
/**
 * The ListLayout class is an implementation of a virtualizer {@link Layout}
 * it is used for creating lists and lists with indented sub-lists.
 *
 * To configure a ListLayout, you can use the properties to define the
 * layouts and/or use the method for defining indentation.
 * The {@link ListKeyboardDelegate} extends the existing virtualizer
 * delegate with an additional method to do this (it uses the same delegate object as
 * the virtualizer itself).
 */
export class ListLayout<T> extends Layout<Node<T>, ListLayoutProps> implements DropTargetDelegate {
    protected rowHeight: number;
    protected estimatedRowHeight: number;
    protected headingHeight: number;
    protected estimatedHeadingHeight: number;
    protected forceSectionHeaders: boolean;
    protected padding: number;
    protected indentationForItem?: (collection: Collection<Node<T>>, key: Key) => number;
    protected layoutInfos: Map<Key, LayoutInfo>;
    protected layoutNodes: Map<Key, LayoutNode>;
    protected contentSize: Size;
    protected collection: Collection<Node<T>>;
    protected isLoading: boolean;
    protected lastWidth: number;
    protected lastCollection: Collection<Node<T>>;
    protected rootNodes: LayoutNode[];
    protected invalidateEverything: boolean;
    protected loaderHeight: number;
    protected placeholderHeight: number;
    protected enableEmptyState: boolean;
    /** The rectangle containing currently valid layout infos. */
    protected validRect: Rect;
    /** The rectangle of requested layout infos so far. */
    protected requestedRect: Rect;
    /**
     * Creates a new ListLayout with options. See the list of properties below for a description
     * of the options that can be provided.
     */
    constructor(options?: ListLayoutOptions<T>);
    getLayoutInfo(key: Key): LayoutInfo;
    getVisibleLayoutInfos(rect: Rect): LayoutInfo[];
    protected layoutIfNeeded(rect: Rect): void;
    protected shouldInvalidateEverything(invalidationContext: InvalidationContext<ListLayoutProps>): boolean;
    validate(invalidationContext: InvalidationContext<ListLayoutProps>): void;
    protected buildCollection(): LayoutNode[];
    protected isValid(node: Node<T>, y: number): boolean;
    protected buildChild(node: Node<T>, x: number, y: number, parentKey: Key | null): LayoutNode;
    protected buildNode(node: Node<T>, x: number, y: number): LayoutNode;
    updateItemSize(key: Key, size: Size): boolean;
    getContentSize(): Size;
    getDropTargetFromPoint(x: number, y: number, isValidDropTarget: (target: DropTarget) => boolean): DropTarget;
}
export interface TableLayoutOptions<T> extends ListLayoutOptions<T> {
    scrollContainer?: 'table' | 'body';
}
export interface TableLayoutProps extends ListLayoutProps {
    columnWidths?: Map<Key, number>;
}
export class TableLayout<T> extends ListLayout<T> {
    collection: TableCollection<T>;
    lastCollection: TableCollection<T>;
    columnWidths: Map<Key, number>;
    stickyColumnIndices: number[];
    isLoading: boolean;
    lastPersistedKeys: Set<Key>;
    persistedIndices: Map<Key, number[]>;
    scrollContainer: 'table' | 'body';
    constructor(options: TableLayoutOptions<T>);
    validate(invalidationContext: InvalidationContext<TableLayoutProps>): void;
    protected buildCollection(): LayoutNode[];
    protected buildNode(node: GridNode<T>, x: number, y: number): LayoutNode;
    getVisibleLayoutInfos(rect: Rect): LayoutInfo[];
    getDropTargetFromPoint(x: number, y: number, isValidDropTarget: (target: DropTarget) => boolean): DropTarget;
}

//# sourceMappingURL=types.d.ts.map
